unit Bitmaps;
{ Bitmaps.tpu
  XPERT Software Productions
  (C) 2000 - Tamer Fakhoury

  FIX ME : BSetPalette does not work on 16-color modes

  UNDER CONSTRUCTION : _2Bit_Display

  LIMITATIONS : Image Size < 65K
}

interface

uses crt,graph;

type { data types }
     _bit     = $00000000..$00000001;
     _byte    = $00000000..$000000FF;
     _word    = $00000000..$0000FFFF;
     _dword   = $00000000..$7FFFFFFF;
     _long    = $80000000..$7FFFFFFF;

     bfHeader = record
                 bfType:_word;
                 bfSize:_dword;
                 bfReserved1:_word;
                 bfReserved2:_word;
                 bfOffBits:_dword;
                end;
     biHeader = record
                 biSize:_dword;
                 biWidth:_long;
                 biHeight:_long;
                 biPlanes:_word;
                 biBitCount:_word;
                 biCompression:_dword;
                 biSizeImage:_dword;
                 biXPixelsPerMeter:_long;
                 biYPixelsPerMeter:_long;
                 biClrUsed:_dword;
                 biClrImportant:_dword;
               end;

     rgbQuad  = record
                 rgbBlue,
                 rgbGreen,
                 rgbRed,
                 rgbReserved:_byte;
                end;

     pBitmap  = ^Bitmap;

     Bitmap   = object
                 { methods }
                 constructor BOpen(szfilename:string);
                 destructor BClose;

                 function BGetWidth:_long;
                 function BGetHeight:_long;
                 function BGetColors:_long;
                 function BGetPalette:boolean;
                 function BGetBits:_word;

                 procedure BSetPalette(rValue_On, gValue_On, bValue_On:boolean);
                 procedure BSetTransparency(On:Boolean; Value:_byte);

                 procedure BDisplay(pX, pY, address:_word);

                 { methods }
                 private
                  { palette routines for 16,24,32-bit bitmaps }
                  function _rgb(r,g,b:_byte):_word;
                  procedure _setrgbpalette(r,g,b:_byte; color:_word);
                  procedure _getrgbpalette(var r,g,b:_byte; color:_word);

                  { bit routines for monochrome bitmaps }
                  function _getbits(var data; index:_byte):_bit;

                  { display routines }
                  procedure _1bit_display(pX, pY, address:_word);
                  procedure _2bit_display(pX, pY, address:_word);
                  procedure _4bit_display(pX, pY, address:_word);
                  procedure _8bit_display(pX, pY, address:_word);
                  procedure _24bit_display(pX, pY, address:_word);

                 { attributes }
                 private
                  bh:file;       { file handle }
                  bf:bfHeader;   { file header }
                  bi:biHeader;   { information header }
                  tColor:integer; { index to transparency color in palette (-1 if none) }
                  palette:array[0..255] of rgbQuad; { bitmap palette }
                  image:pointer;    { bitmap bits }
                end;

implementation

constructor Bitmap.BOpen(szfilename:string);
begin
 assign(bh, szfilename);
 reset(bh,1);
 blockread(bh, bf, sizeof(bfHeader));
 blockread(bh, bi, sizeof(biHeader));
 image:=nil;
 tColor:=-1; { disable transparency }
end;

destructor Bitmap.BClose;
begin
 close(bh);
 if (image <> nil)
  then freemem(image, bi.biSizeImage);
end;

function Bitmap.BGetWidth:_long;
begin
 BGetWidth:=bi.biWidth;
end;

function Bitmap.BGetHeight:_long;
begin
 BGetHeight:=bi.biHeight;
end;

function Bitmap.BGetColors:_long;
begin
 BGetColors:=_long(1 shl _long( bi.biBitCount ) );
end;

function Bitmap.BGetPalette:boolean;
begin
 if bi.biBitCount > 8 then begin
  BGetPalette:=false; { no palette }
  exit;
 end;
 blockread(bh, palette, (1 shl bi.biBitCount) * sizeof(rgbQuad));
 BGetPalette:=true; { palette found }
end;

function Bitmap.BGetBits:_word;
begin
 if (bi.biSizeImage > $ffff)
    then exit; { > 65K }
 getmem(image, bi.biSizeImage);
 seek(bh, bf.bfOffBits);
 blockread(bh, image^, bi.biSizeImage);
 BGetBits:=seg(image^); { return address of image bits }
end;

procedure Bitmap.BSetPalette(rValue_On, gValue_On, bValue_On:boolean);
var index,
    colors:_byte;
begin
 colors:=(1 shl _byte( bi.biBitCount ) );
 for index:=0 to colors-1 do begin
  port[$3c8]:=index;
  if rValue_On
   then port[$3c9]:=palette[index].rgbRed shr 2
   else port[$3c9]:=0;
  if gValue_On
   then port[$3c9]:=palette[index].rgbGreen shr 2
   else port[$3c9]:=0;
  if bValue_On
   then port[$3c9]:=palette[index].rgbBlue shr 2
   else port[$3c9]:=0;
 end;
end;

procedure Bitmap.BSetTransparency(On:Boolean; Value:_byte);
begin
 if On and (bi.biBitCount <= 8) { palette present }
  then tColor:=Value
  else tColor:=-1;
end;

procedure Bitmap.BDisplay(pX, pY, address:_word);
begin
 case bi.biBitCount of
   1:_1bit_display(pX, pY, address);
   2:_2bit_display(pX, pY, address);
   4:_4bit_display(pX, pY, address);
   8:_8bit_display(pX, pY, address);
  24:_24bit_display(pX, pY, address);
 end;
end;

{ Private methods }

function Bitmap._rgb(r,g,b:_byte):_word;
begin
 _rgb :=_word( (r) or (g shl 4) or (b shl 8) );
end;

procedure Bitmap._setrgbpalette(r,g,b:_byte; color:_word);
begin
 portw[$3c8]:=color;
 portw[$3c9]:=r;
 portw[$3c9]:=g;
 portw[$3c9]:=b;
end;

procedure Bitmap._getrgbpalette(var r,g,b:_byte; color:_word);
begin
 portw[$3c7]:=color;
 r:=_byte( portw[$3c9] );
 g:=_byte( portw[$3c9] );
 b:=_byte( portw[$3c9] );
end;

function Bitmap._getbits(var data; index:_byte):_bit;
var bits:_byte absolute data;
begin
 if ( (bits and (1 shl index) ) <> 0 )
  then _getbits:=1
  else _getbits:=0;
end;

procedure Bitmap._1bit_display(pX, pY, address:_word);
var compressed_width,
    x, y, i:_word;
    pColor:_byte;
begin
 compressed_width := (bi.biSizeImage div bi.biHeight);
 for y:=0 to bi.biHeight - 1 do
  for x:=0 to (bi.biWidth div 8) do begin { ignore padded bits }
   pColor:=mem[address:y * ( compressed_width ) + x];
   for i:=0 to 7 do { scan through bits }
    putpixel(pX + ( (x + 1) shl 3 ) - i, pY + ( bi.biHeight - (y + 1) ), _getbits(pColor, i) );
  end;
end;

procedure Bitmap._2bit_display(pX, pY, address:_word);
begin
 { under construction }
end;

procedure Bitmap._4bit_display(pX, pY, address:_word);
var extra_bits,
    x, y:_word;
    pColor:_byte;
begin
 { formula to calculate padded bits... }
 extra_bits:=( bi.biSizeImage - ( (bi.biWidth div 2) * bi.biHeight ) ) div bi.biHeight * 2;
 for y:=0 to bi.biHeight - 2 do
  for x:=0 to (bi.biWidth shr 1) - 1 do begin
   pColor:=mem[address:y * ( (bi.biWidth + extra_bits) shr 1 ) + x] div 16; { lower bit pixel }
   if (pColor <> tColor) { transparent ? }
    then putpixel(pX + (x shl 1), pY + ( bi.biHeight - (y + 1) ), pColor);
   { bitmaps are stored upside down... }
   pColor:=mem[address:y * ( (bi.biWidth + extra_bits) shr 1 ) + x] mod 16; { upper bit pixel }
   if (pColor <> tColor) { transparent ? }
    then putpixel(pX + (x shl 1 + 1), pY + ( bi.biHeight - (y + 1) ), pColor);
   { bitmaps are stored upside down... }
  end;
end;

procedure Bitmap._8bit_display(pX, pY, address:_word);
type l = 0..15;
var extra_bits,
    x, y:_word;
    pColor:_byte;
begin
 { formula to calculate padded bits... }
 extra_bits:=( bi.biSizeImage - (bi.biWidth * bi.biHeight) ) div bi.biHeight;
 for y:=0 to bi.biHeight - 1 do
  for x:=0 to bi.biWidth - 1 do begin
   pColor:=mem[address:y * (bi.biWidth + extra_bits) + x];
   if ( getmaxColor = 15) then
     pColor:=pColor and $0f; { map onto a 16 color palette }
   if (pColor <> tColor) { tarnsparent ? }
    then putpixel(pX + x, pY + (bi.biHeight - (y + 1) ), pColor); { display pixel onto screen }
   { bitmaps are stored upside down... }
  end;
end;

procedure Bitmap._24bit_display(pX, pY, address:_word);
var extra_bits,
    x, y:_word;
    rgbi,r,g,b:_byte;
    rgb:array[0..2] of _byte;
    pColor:integer;
begin
 { formula to calculate padded bits...}
 extra_bits:=(bi.biSizeImage - (bi.biWidth * 3 * bi.biHeight) ) div bi.biHeight;
 for y:=0 to bi.biHeight-1 do
  for x:=0 to bi.biWidth-1 do begin
   { read rgb triplette }
   for rgbi:=0 to 2 do
    rgb[rgbi]:=mem[address:( y * ( (bi.biWidth * 3) + extra_bits ) ) + ( (x * 3) + (rgbi) )];
   pColor:=_rgb(rgb[2] , rgb[1] , rgb[0]);
   case getmaxColor of
     15:pColor:=(pColor and $00ff) shr 4; { map onto a 16 color palette }
    255:pColor:=(pColor and $ffff) shr 8; { map onto a 256 color palette }
   end;
   _getrgbpalette(r, g, b, pColor);
   { make sure you set rgb for each color only once }
   if ( (r <> rgb[2] shr 2) or (g <> rgb[1] shr 2) or (b <> rgb[0] shr 2) ) then
    _setrgbpalette(rgb[2] shr 2, rgb[1] shr 2, rgb[0] shr 2, pColor);
   { bitmaps are stored upside down...}
   putpixel(pX + x, pY + (bi.biHeight - (y + 1)), pColor);
 end;
end;

begin
end.